#include "classes.h"
#include <ctype.h>

extern int num_rules, delimiter, do_uncompress;

rules::rules() 
{
  num_rules ++;
  guard = new symbols(ulong(this));
  guard->point_to_self();
  count = number = Usage = 0;
}

rules::~rules() { 
  num_rules --;
  delete guard;
}

symbols *rules::first() { return guard->next(); }
symbols *rules::last() { return guard->prev(); }

int symbols::check() 
{
  if (is_guard() || n->is_guard()) return 0;
  
  symbols **x = find_digram(this);
  if (!x) return 0;
  symbols *y[100];
  int i;

  for (i = 0; i < K; i ++) y[i] = x[i];
  
  for (i = 0; i < K; i ++)
    if (int(x[i]) <= 1) {
      x[i] = this;
      return 0;
    }
  
  for (i = 0; i < K; i ++) 
    if (x[i]->next() == this || next() == x[i])
      return 0;
  
  rules *r;
  
  // reuse an existing rule
  
  for (i = 0; i < K; i ++)
    if (y[i]->prev()->is_guard() && y[i]->next()->next()->is_guard()) {
      r = y[i]->prev()->rule();
      substitute(r); 
      
      // check for an underused rule
      
      if (r->first()->nt() && r->first()->rule()->freq() == 1) r->first()->expand();
      // if (r->last()->nt() && r->last()->rule()->freq() == 1) r->last()->expand();
      return 1;
    }

  // create a new rule
  
  r = new rules;    
  
  r->last()->insert_after(new symbols(value()));
  r->last()->insert_after(new symbols(next()->value()));
  
  x[0] = r->first();
  
  for (i = 0; i < K; i ++) {
    y[i]->substitute(r);
    y[i] = (symbols *) 1; // should be x
  }
  
  substitute(r);
  
  // check for an underused rule
  
  if (r->first()->nt() && r->first()->rule()->freq() == 1) r->first()->expand();
  //  if (r->last()->nt() && r->last()->rule()->freq() == 1) r->last()->expand();
  
  return 1;
}

// This symbol is the last reference to its rule. It is deleted, and the
// contents of the rule substituted in its place.

void symbols::expand() {
  symbols *left = prev();
  symbols *right = next();
  symbols *f = rule()->first();
  symbols *l = rule()->last();

  symbols **m = find_digram(this);
  if (!m) return;
  delete rule();
  
  for (int i = 0; i < K; i ++)
    if (m[i] == this) m[i] = (symbols *) 1;
  s = 0; // if we don't do this, deleting the symbol tries to deuse the rule!
  delete this;
  
  join(left, f);
  join(l, right);
}

// Replace a digram with a non-terminal

void symbols::substitute(rules *r)
{
  symbols *q = p;
  
  delete q->next();
  delete q->next();

  q->insert_after(new symbols((ulong) r));
  
  if (!q->check()) q->n->check();
}

// pick a prime! (large enough to hold every digram in the grammar, with room
// to spare

#define PRIME 4265561
//#define PRIME 62265551

// Standard open addressing or double hashing. See Knuth.

#define HASH(one, two) (((one) << 16 | (two)) % PRIME)
#define HASH2(one) (17 - ((one) % 17))

symbols **table = 0;

symbols **find_digram(symbols *s)
{
  if (!table) {
    table = (symbols **) malloc(PRIME * K * sizeof(symbols *));
    memset(table, 0, PRIME * K * sizeof(symbols *));
  }

  ulong one = s->value();
  ulong two = s->next()->value();

  if (one == delimiter || two == delimiter) return 0;

  int jump = HASH2(one) * K;
  int insert = -1;
  int i = HASH(one, two) * K;

  while (1) {
    symbols *m = table[i];
    if (!m) {
      if (insert == -1) insert = i;
      return &table[insert];
    } 
    else if (int(m) == 1) insert = i;
    else if (m->value() == one && m->next()->value() == two) return &table[i];
    i = (i + jump) % PRIME;
  }
}

void rules::reproduce()
{
  for (symbols *p = first(); !p->is_guard(); p = p->next())
    p->reproduce();
}

ostream &operator << (ostream &o, symbols &s)
{
  extern int numbers;

  if (s.nt()) {
    //    assert(s.rule()->index() != 9999);
    //    assert(s.rule()->index() < current_rule);
    o << s.rule()->index();
  }
  else if (numbers) o << '&' << s.value();
  else if (do_uncompress) o << char(s.value());
  else if (s.value() == '\n') o << "\\n";
  else if (s.value() == '\t') o << "\\t";
  else if (s.value() == ' ' ) o << '_';
  else if (s.value() == '\\' ||
	   s.value() == '(' ||
	   s.value() == ')' ||
	   s.value() == '_' ||
	   isdigit(s.value()))
    o << '\\' << char(s.value());
  else o << char(s.value());

  return o;
}

void rules::output() 
{
  symbols *s;
  
  for (s = first(); !s->is_guard(); s = s->next())
    if (s->nt() && s->rule()->index() == 0) 
      s->rule()->output();

  number = current_rule ++;

  for (s = first(); !s->is_guard(); s = s->next())
    cout << *s << ' ';

  extern int print_rule_usage;
  if (print_rule_usage) cout << "\t(" << Usage << ")";
  
  cout << endl;
}

