#include <assert.h>
#include <fstream.h>
#include <memory.h> // for memset
#include <stdlib.h> // for malloc

extern int num_symbols, current_rule, K;

class symbols;
class rules;
ostream &operator << (ostream &o, symbols &s);

typedef unsigned long ulong;

// This finds a matching digram in the grammar, and returns a pointer 
// to its entry in the hash table so it can be changed or deleted as necessary

extern symbols **find_digram(symbols *s);

///////////////////////////////////////////////////////////////////////////

class rules {
  // the guard node in the linked list of symbols that make up the rule
  // It points forward to the first symbol in the rule, and backwards
  // to the last symbol in the rule. Its own value points to the rule data 
  // structure, so that symbols can find out which rule they're in
  
  symbols *guard;     
  
  //  count keeps track of the number of times the rule is used in the grammar
  int count;

int Usage;

  // this is just for numbering the rules nicely for printing; it's
  // not essential for the algorithm
  int number;

public:
  void send();
  void send_pointer();
  int compressed_length();
  void output();
  void output2();

  rules();
  ~rules();

  void reuse() { count ++; }
  void deuse() { count --; };

  symbols *first();
  symbols *last();
  
  int freq()  { return count; };
  int usage()  { return Usage; };
  void usage(int i)  { Usage += i; };
  int index()  { return number; };
  void index(int i)  { number = i; };
  void reproduce();
};

class symbols {
  symbols *n, *p;
  ulong s;

public:

  int expansion_length() { return 1; }; /////////////////////////
  int mint() { return 1; }; /////////////////////////
  int sint() { return 1; }; /////////////////////////
  int prefix() { return 1; }; /////////////////////////
  symbols *next_same() { return this; }; /////////////////////////

  // initializes a new symbol. If it is non-terminal, increments the reference
  // count of the corresponding rule

  void reproduce() {
    if (nt()) rule()->reproduce();
    else cout << *this;
  }


  symbols(ulong sym) {
    s = sym;
    p = n = 0;
    if (nt()) rule()->reuse();
    num_symbols ++;
  }

  // links two symbols together, removing any old digram from the hash table

  static void join(symbols *left, symbols *right) {
    if (left->n) left->delete_digram();
    left->n = right; right->p = left;
  }

  // cleans up for symbol deletion: removes hash table entry and decrements
  // rule reference count

  ~symbols() {
    join(p, n);
    if (!is_guard()) {
      delete_digram();
      if (nt()) rule()->deuse(); 
    }
    num_symbols --;
  }

  // inserts a symbol after this one.

  void insert_after(symbols *y) {
    join(y, n);
    join(this, y);
  };

  // removes the digram from the hash table

  void delete_digram() {
    if (is_guard() || n->is_guard()) return;
    symbols **m = find_digram(this);
    if (m == 0) return;
    for (int i = 0; i < K; i ++) if (m[i] == this) m[i] = (symbols *) 1;
  }
  
  // returns true if this is the guard node marking the beginning/end of a rule

  int is_guard() { return nt() && rule()->first()->prev() == this; };

  // nt() returns true if a symbol is non-terminal. The number determines the
  // alphabet size. Anything below this is a terminal, above this is a
  // non-terminal. For ascii text, it should be 255, but for, say, word-based
  // sequences, it should be much higher.  It just has to be big enough; there
  // is no penalty for it being too large

  int nt() { return s >= 20000000;};

  symbols *next() { return n;};
  symbols *prev() { return p;};
  inline ulong value() { return s;};
  
  // assuming this is a non-terminal, rule() returns the corresponding rule

  rules *rule() { return (rules *) s;};

  void substitute(rules *r);
  static void match(symbols *s, symbols *m);
  
  // checks a new digram. If it appears elsewhere, deals with it by calling
  // match(), otherwise inserts it into the hash table

  int check();
  
  void expand();

  void point_to_self() { join(this, this); };

  void send();
};

