#include <assert.h>
#include <fstream.h>
#include <stdio.h>
#include <math.h>

#include "classes.h"

extern "C" {
#include "arith.h"
#include "stats.h"
#include "bitio.h"
}

static context *symbol, *lengths, *keep;

extern int compress;

#define START_RULE 256
#define END_OF_FILE 257
#define STOP_FORGETTING 258
#define FIRST_RULE 259

#define MAX_LENGTH 100

void start_compress()
{
  int i;

  keep = create_context(2, STATIC);
  install_symbol(keep, 0);
  install_symbol(keep, 1);

  symbol = create_context(FIRST_RULE, DYNAMIC);

  for (i = 0; i < 256; i++) install_symbol(symbol, i);

  install_symbol(symbol, START_RULE);
  install_symbol(symbol, STOP_FORGETTING);
  install_symbol(symbol, END_OF_FILE);

  lengths = create_context(MAX_LENGTH + 1, STATIC);
  for (i = 0; i < MAX_LENGTH + 1; i++) install_symbol(lengths, i);

  if (compress) {
    startoutputtingbits();
    start_encode();
  }
  else {
    startinputtingbits();
    start_decode();
  }
}     

static int forgetting = 1;

void stop_forgetting() 
{
  encode(symbol, STOP_FORGETTING);
  forgetting = 0;
}

void end_compress() {
  if (compress) {
    encode(symbol, END_OF_FILE);
    finish_encode();
  }
  else {
    finish_decode();
    doneinputtingbits();
  }
}

int current_rule = FIRST_RULE;

void encode_rule(rules *r, int keepi) 
{
  encode(symbol, r->index());
  if (keepi < 2 && forgetting) {
    encode(keep, keepi);
    if (!keepi) delete_symbol(symbol, r->index());
  }
}

void rules::output2() 
{
  symbols *s;
  
  number = current_rule ++;

  encode(symbol, START_RULE);
  install_symbol(symbol, number);

  int l = 0;
  for (s = first(); !s->is_guard(); s = s->next()) l ++;

  if (l >= MAX_LENGTH) {
    encode(lengths, MAX_LENGTH);
    arithmetic_encode(l, l + 1, 10000);
  }
  else encode(lengths, l);
  
  for (s = first(); !s->is_guard(); s = s->next())
    if (s->nt() && s->rule()->index() == 0) s->rule()->output2();
    else if (s->nt()) encode_rule(s->rule(), 2);
    else encode(symbol, s->value());
}

void forget(symbols *s) 
{
  rules *r = 0;

  if (s->nt()) {
    r = s->rule();
    delete s;

    if (r->freq() > 0) {
      if (r->index() == 0) r->output2();
      else encode_rule(r, 1);
    }
    else {
      encode_rule(r, 0);

      while (r->first()->next() != r->first()) delete r->first();
      delete r;
    }
  }
  else {
    encode(symbol, s->value());
    delete s;
  }
}

static rules **R;

int get_symbol() 
{
  int i = decode(symbol);

  if (i == START_RULE) {
    int n = current_rule ++;
    assert(current_rule < 100000);
    
    R[n] = new rules;
    install_symbol(symbol, n);
    int l = decode(lengths);

    if (l == MAX_LENGTH) {
      l = arithmetic_decode_target(10000);
      arithmetic_decode(l, l + 1, 10000);
    }
    
    for (int j = 0; j < l; j ++) {
      ulong x = get_symbol();
      if (x >= FIRST_RULE) x = ulong(R[x]);
	R[n]->last()->insert_after(new symbols(x));
    }
    return n;
  }
  else return i;
}
  
void uncompress() 
{
  R = (rules **) malloc(100000 * sizeof(rules *));

  start_compress();

  while (1) {
    int current = current_rule;

    int i = get_symbol();
    
    if (i == END_OF_FILE) break;
    else if (i == STOP_FORGETTING) forgetting = 0;
    else if (i >= FIRST_RULE) {
      R[i]->reproduce();
      if (i < current && forgetting) {
	int keepi = decode(keep);
	
	if (!keepi) {
	  delete_symbol(symbol, i);
	  while (R[i]->first()->next() != R[i]->first()) delete R[i]->first();
	  delete R[i];
	}
      }
    }
    else cout << char(i);
  }
  
  end_compress();
}
